local MAJOR = "LibScriptableDisplayHook-1.0"
local MINOR = 14

local LibHook = LibStub:NewLibrary(MAJOR, MINOR)
if not LibHook then return end

local hooks = {}

-- Create a hook
-- @param obj The object to hook
-- @param method The object's method to hook
-- @param handler This is your handler function
-- @param data Some data to pass back to the handler
-- @param secure Whether this is a secure hook or not
-- @return Nothing
function LibHook:CreateHook(obj, method, target, secure)
	assert(type(obj) == "table", "CreateHook requires an object to hook")
	assert(type(method) == "string", "CreateHook requires a string representation of the method to hook")
	assert(type(target) == "table", "CreateHook requires a target to hook")
	assert(type(target[method]) == "function", "Target method is not a function")
	local handler = target[method]
	hooks[obj] = hooks[obj] or {}
	hooks[obj][method] = {obj[method]}
	obj[method] = function(...)
		if hooks[obj][method][2] then
			return hooks[obj][method][1](...)
		end
		if true or secure then
			handler(target, ...)
			return hooks[obj][method][1](...)
		else
			return handler(target, ...)
		end
	end
end

-- Unhook a hook
-- @param obj The object to unhook
-- @param method The method to unhook
-- @return Nothing
function LibHook:Unhook(obj, method)
	hooks[obj][method][2] = true
end
